# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .plugins_base import Plugin
from .capture.plugin_capture import LedgerCapturePlugin
